def eval_sort(statements: dict, names: set):
    """Figures out the order in which statements should be evaluated"""
    names = set(names)
    dependencies = {k: v.required for k, v in statements.items()}
    result = []
    delta = set(dependencies).intersection(names)
    if len(delta) != 0:
        raise ValueError(f"names and statements overlap: {', '.join(delta)}")

    while True:
        transaction = []
        for name, depends_on in dependencies.items():
            if len(depends_on - names) == 0:
                transaction.append(name)
                names.add(name)
                result.append(statements[name])
        if len(transaction) == 0:
            if len(dependencies) > 0:
                info = []
                for k, v in dependencies.items():
                    info.append(f"{statements[k]}: missing {', '.join(map(repr, set(v) - names))}")
                info = "\n".join(info)
                raise ValueError(
                    f"{len(dependencies)} expressions cannot be evaluated:\n{info}")
            return result
        else:
            for i in transaction:
                del dependencies[i]


def eval_all(statements: list, names: dict):
    """Evaluates all expressions from the dict"""
    result = []
    names = names.copy()
    for statement in statements:
        v = statement.eval(names)
        names[statement.name] = v
        result.append(v)
    return result



from types import FunctionType

from .parsing import split_assignment, iter_template_blocks


class EvalBlock:
    def __init__(self, code, name, fmt):
        """Represents a statement that can be evaluated"""
        self.code = code
        self.name = name
        self.fmt = fmt

    @classmethod
    def from_string(cls, text, defined_file="unknown", defined_line="u", defined_char="u"):
        name, fmt, text = split_assignment(text)
        if name is None:
            defined_file = ''.join(i if i.isalnum() else "_" for i in defined_file)
            name = f"anonymous_{defined_file}_l{defined_line}c{defined_char}"
        code = compile(text, "<string>", "eval")
        return cls(code, name, fmt)

    @property
    def required(self):
        return set(self.code.co_names)

    def names_missing(self, names):
        return set(self.required) - set(names)

    def eval(self, names):
        delta = self.names_missing(names)
        if len(delta) > 0:
            raise ValueError(f"missing following names: {', '.join(map(repr, delta))}")
        func = FunctionType(self.code, names)
        return func()

    def format_value(self, val):
        if self.fmt == "suppress":
            return ""
        elif self.fmt is None:
            return str(val)
        else:
            return f"{val:{self.fmt}}"

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"EvalBlock('{self.__str__()}')"


class Template:
    def __init__(self, name, chunks):
        """A file with multiple statements to evaluate"""
        self.name = name
        self.chunks = chunks

    @classmethod
    def from_text(cls, name, text):
        itr = iter_template_blocks(text)
        chunks = []
        for pos, i in itr:
            chunks.append(i)  # regular text block
            try:
                pos, i = next(itr)  # eval block
                chunks.append(EvalBlock.from_string(
                    i,
                    defined_file=name,
                    defined_line=text[:pos].count("\n") + 1,
                    defined_char=pos - text[:pos].rfind("\n"),
                ))
            except StopIteration:
                pass
        return cls(name, chunks)

    @classmethod
    def from_file(cls, f):
        return cls.from_text(f.name, f.read())

    def write(self, stack, f):
        for chunk in self.chunks:
            if isinstance(chunk, str):
                f.write(chunk)
            elif isinstance(chunk, EvalBlock):
                f.write(chunk.format_value(stack[chunk.name]))
            else:
                raise NotImplementedError(f"unknown {chunk=}")

    def is_trivial(self):
        for chunk in self.chunks:
            if not isinstance(chunk, str):
                return False
        return True

    def __repr__(self):
        return f"GridFile(name={repr(self.name)}, chunks=[{len(self.chunks)} chunks])"


def variable_list_template(variable_names, name=".variables"):
    """Constructs a template with variable names"""
    # TODO: fix a hack here var = var?
    return Template.from_text(name, '\n'.join(f"{i} = {{% {i} = {i} %}}" for i in variable_names))



import re


def repr_pos(text, pos, left=16, width=32, placeholder="..."):
    """Represent position in text"""
    if pos >= len(text):
        raise ValueError(f"invalid {pos=} for text of length {len(text)}")
    start = text.rfind("\n", 0, pos) + 1
    end = text.find("\n", pos + 1)
    if end == -1:
        end = len(text)
    repr_chunk = text[start:end]
    pos -= start

    if repr_chunk[pos] == "\n":
        two_lines = repr_chunk.split("\n")
        assert len(two_lines) == 2
        two_lines = list(
            "(empty line)" if len(i) == 0 else i if len(i) <= width else f"{i[:width]}{placeholder}"
            for i in two_lines
        )
        return f"Line break between the following lines:\n" + "\n".join(two_lines)

    else:
        if pos > left:
            repr_chunk = f"{placeholder}{repr_chunk[pos - left:]}"
            pos = left + len(placeholder)
            width += len(placeholder)
        if len(repr_chunk) > width:
            repr_chunk = f"{repr_chunk[:width]}{placeholder}"
        return f"{repr_chunk}\n" + ' ' * pos + "^"


def iter_template_blocks(text, left="{%", right="%}", escape="\\"):
    """Iterates over template blocks"""
    len_esc = len(escape)

    def _is_escaped_at(_pos):
        return _pos > len_esc and text[_pos - len_esc:_pos] == escape

    pos = 0

    def _iter_until(_pattern):
        nonlocal pos
        len_pattern = len(_pattern)
        while True:
            pattern_start = text.find(_pattern, pos)
            if pattern_start < 0:
                yield pos, text[pos:]
                pos = len(text)
                yield pos, None  # indicates end of file without finding the pattern
                break
            elif _is_escaped_at(pattern_start):
                yield pos, text[pos:pattern_start - len_esc]
                yield pattern_start, _pattern
                pos = pattern_start + len_pattern
            else:
                yield pos, text[pos:pattern_start]
                pos = pattern_start + len_pattern
                break

    while True:
        # find left
        pieces = list(_iter_until(left))
        if pieces[-1][1] is None:  # nothing left
            yield pieces[0][0], ''.join(i for _, i in pieces[:-1])
            break
        else:
            yield pieces[0][0], ''.join(i for _, i in pieces)
        start = pos

        # find right
        pieces = list(_iter_until(right))
        if pieces[-1][1] is None:  # did not find end of template
            raise ValueError(f"missing closing bracket for template starting at {start}:\n{repr_pos(text, start)}")
        else:
            yield pieces[0][0], ''.join(i for _, i in pieces)


def split_assignment(text, name_pattern=re.compile(r"^\s*(?P<name>[\w_]+)\s*="),
                     format_pattern=re.compile(r"\:(?P<format>[\w_\.\s]+)$")):
    name_match = re.search(name_pattern, text)
    if name_match is None:
        name = None
        start = 0
    else:
        name = name_match.group("name")
        if name[0] in "0123456789":
            raise ValueError(f"invalid variable name: {name}")
        start = name_match.end()

    format_match = re.search(format_pattern, text)
    if format_match is None:
        fmt = None
        end = len(text)
    else:
        fmt = format_match.group("format").rstrip()
        end = format_match.start()
    return name, fmt, text[start:end].strip()



from pathlib import Path
from functools import partial
import shutil

from .template import Template


def match_files(patterns, root=".", apply=None, exclude=None, recursive=False, hidden=False, allow_empty=False):
    """Collects files matching a list of patterns"""
    if len(patterns) == 0 and not allow_empty:
        raise ValueError(f"no patterns or files provided")
    root = Path(root)
    if exclude is None:
        exclude = set()

    result = []
    matched_so_far = set()
    for pattern in patterns:
        anything_matched = False
        matched_total = 0
        matched_files = 0
        for match in (root.rglob if recursive else root.glob)(pattern):
            matched_total += 1
            if match.is_file() and match not in matched_so_far and (hidden or not match.name.startswith(".")):
                matched_so_far.add(match)
                matched_files += 1
                if match not in exclude:
                    if apply is not None:
                        match = apply(match)
                        if match is not None:
                            result.append(match)
                            anything_matched = True
                    else:
                        result.append(match)
                        anything_matched = True

        if not anything_matched:
            raise ValueError(f"pattern '{pattern}' in '{str(root)}' matched 0 files (matched total: {matched_total}, "
                             f"files: {matched_files})")
    return result


def _maybe_template(candidate):
    with open(candidate, 'r') as f:
        result = Template.from_file(f)
    if result.is_trivial():
        return
    return result


match_template_files = partial(match_files, apply=_maybe_template)


def write_grid(directory_name, stack, files_static, files_grid, root):
    """Writes grid folder contents"""
    root = Path(root)

    def _translate_path(p):
        return root / directory_name / Path(p).relative_to(root)

    for src in files_static:
        dst = _translate_path(src)
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(src, dst)

    for f in files_grid:
        src = f.name
        dst = _translate_path(src)
        dst.parent.mkdir(parents=True, exist_ok=True)
        with open(dst, "w") as ff:
            f.write(stack, ff)
        try:
            shutil.copystat(src, dst)
        except FileNotFoundError:
            pass  # virtual file



import argparse
import json
import logging
import shutil
import subprocess
import sys
from pathlib import Path
from functools import reduce
from operator import mul
from warnings import warn

from .algorithm import eval_sort, eval_all
from .tools import combinations
from .template import EvalBlock, variable_list_template
from .grid_builtins import builtins
from .files import match_files, match_template_files, write_grid

arg_parser = argparse.ArgumentParser(description="Creates arrays [grids] of similar files and folders")
arg_parser.add_argument("-f", "--files", nargs="+", help="files to be processed", metavar="FILE", default=tuple())
arg_parser.add_argument("-t", "--static", nargs="+", help="files to be copied", metavar="FILE", default=tuple())
arg_parser.add_argument("-r", "--recursive", help="visit sub-folders when matching file names", action="store_true")
arg_parser.add_argument("-n", "--name", help="grid folder naming pattern", metavar="PATTERN", default="grid%d")
arg_parser.add_argument("-m", "--max", help="maximum allowed grid size", metavar="N", default=10_000)
arg_parser.add_argument("-s", "--state", help="state file name", metavar="FILE", default=".grid")
arg_parser.add_argument("-l", "--log", help="log file name", metavar="FILE", default=".grid.log")
arg_parser.add_argument("--root", help="root folder for scanning/placing grid files", default=".")
arg_parser.add_argument("action", help="action to perform", choices=["new", "run", "cleanup", "distribute"])
arg_parser.add_argument("extra", nargs="*", help="extra action arguments for 'run' and 'distribute'")


class Engine:
    def __init__(self, action, extra, template_files, static_files, root, recursive, name, max_size, state_fn, log_fn):
        self.action = action
        self.extra = extra
        self.template_files = template_files
        self.static_files = static_files
        self.root = root
        self.recursive = recursive
        self.name = name
        self.max_size = max_size
        self.state_fn = state_fn
        self.log_fn = log_fn

    @classmethod
    def from_argparse(cls, options):
        return cls(
            action=options.action,
            extra=options.extra,
            template_files=options.files,
            static_files=options.static,
            root=options.root,
            recursive=options.recursive,
            name=options.name,
            max_size=options.max,
            state_fn=options.state,
            log_fn=options.log,
        )

    def setup_logging(self):
        logging.basicConfig(filename=self.log_fn, filemode="w", level=logging.INFO)

    def load_state(self):
        """Reads the grid state"""
        logging.info(f"Loading grid state from '{self.state_fn}'")
        try:
            with open(self.state_fn, "r") as f:
                return json.load(f)
        except FileNotFoundError as e:
            raise FileNotFoundError(f"Grid file does not exit: {repr(e.filename)}") from e

    def save_state(self, state):
        """Saves the grid state"""
        with open(self.state_fn, "w") as f:
            json.dump(state, f, indent=4)

    def folder_name(self, index):
        """Folder name convention"""
        return self.name % index

    def match_static(self):
        logging.info("Matching static files")
        result = match_files(self.static_files, allow_empty=True, recursive=self.recursive)
        for i in result:
            logging.info(f"  {str(i)}")
        logging.info(f"Total: {len(result)} files")
        return result

    def match_templates(self, exclude):
        logging.info("Matching template files")
        request = []
        if self.action == "distribute":
            request = (*self.template_files, *self.extra)
        elif self.action == "new":
            request = self.template_files
            if len(request) == 0:
                request = "*",
        result = match_template_files(request, recursive=self.recursive, exclude=exclude)
        for i in result:
            logging.info(f"  {str(i)}")
        logging.info(f"Total: {len(result)} files")
        return result

    @staticmethod
    def collect_statements(files_grid):
        statements = {}
        for grid_file in files_grid:
            for chunk in grid_file.chunks:
                if isinstance(chunk, EvalBlock):
                    if chunk.name in statements:
                        raise ValueError(f"duplicate statement {chunk} (also {statements[chunk.name]}")
                    else:
                        statements[chunk.name] = chunk
        return statements

    def group_statements(self, statements):
        statements_core = {}
        statements_dependent = {}

        for name, statement in statements.items():
            logging.info(repr(statement))
            if len(statement.names_missing(builtins)) == 0 and self.action != "distribute":
                logging.info("  core, evaluating ...")
                result = statement.eval(builtins)
                if "__len__" not in dir(result):
                    result = [result]
                logging.info(f"  result: {result} (len={len(result)})")
                statements_core[name] = result

            else:
                logging.info(f"  depends on: {', '.join(map(repr, statement.required))}")
                statements_dependent[name] = statement
        total = reduce(mul, map(len, statements_core.values())) if len(statements_core) else 1
        logging.info(f"Total: {len(statements_core)} core statement(s) ({total} combination(s)), "
                     f"{len(statements_dependent)} dependent statement(s)")
        if total > self.max_size:
            raise RuntimeError(f"the total grid size {total} is above threshold {self.max_size}")
        return statements_core, statements_dependent, total

    def check_folder_conflicts(self, *args):
        for i in range(*args):
            p = Path(self.name % i)
            if p.exists():
                raise RuntimeError(f"file or folder '{str(p)}' already exists")

    def run_new(self, builtins=builtins):
        """
        Performs the new action.

        Creates an array of grid folders.
        """
        logging.info("Creating a new grid")

        files_static = self.match_static()
        files_grid = self.match_templates(files_static)
        statements = self.collect_statements(files_grid)

        reserved_names = set(builtins) | {"__grid_folder_name__", "__grid_id__"}
        overlap = set(statements).intersection(reserved_names)
        if len(overlap) > 0:
            raise ValueError(f"the following names used in the grid are reserved: {', '.join(overlap)}")

        statements_core, statements_dependent, total = self.group_statements(statements)

        # Read previous run
        self.check_folder_conflicts(total)
        grid_state = {"grid": {}, "names": list(statements)}
        index = 0

        if len(statements_core) == 0:
            warn(f"No fixed groups found")

        # Figure out order
        ordered_statements = eval_sort(statements_dependent, reserved_names | set(statements_core))
        # Add variables template file
        files_grid.append(variable_list_template(sorted(statements.keys())))
        # Iterate over possible combinations and write a grid
        for stack in combinations(statements_core):
            scratch = self.folder_name(index)
            stack["__grid_folder_name__"] = scratch
            stack["__grid_id__"] = index

            values = eval_all(ordered_statements, {**stack, **builtins})
            stack.update({statement.name: v for statement, v in zip(ordered_statements, values)})
            grid_state["grid"][scratch] = {"stack": stack}
            logging.info(f"  composing {scratch}")
            write_grid(scratch, stack, files_static, files_grid, self.root)
            index += 1

        # Save state
        self.save_state(grid_state)

    def run_distribute(self):
        """
        Runs grid distribute.

        Distributes one or many files over the existing grid.
        """
        logging.info("Distributing over the existing grid")
        current_state = self.load_state()
        logging.info(f"   grid folders: {len(current_state['grid'])}")

        files_static = self.match_static()
        files_grid = self.match_templates(files_static)
        statements = self.collect_statements(files_grid)

        reserved_names = set(builtins) | {"__grid_folder_name__", "__grid_id__"}
        overlap = set(statements).intersection(reserved_names)
        if len(overlap) > 0:
            raise ValueError(f"the following names used in the grid are reserved: {', '.join(overlap)}")

        statements_core, statements_dependent, total = self.group_statements(statements)
        if len(statements_core) > 0:
            raise ValueError(f"(new) core statements are not allowed when distributing: {', '.join(statements_core)}")
        overlap = set(current_state["names"]).intersection(set(statements))
        if len(overlap) > 0:
            raise ValueError(f"new names overlap with the ones previously defined: {', '.join(overlap)}")
        if len(statements_dependent) == 0:
            warn("No dependent statements found. File(s) will be distributed as-is.")

        logging.info(f"Distributing files into {len(current_state['grid'])} folders")
        exceptions = []

        # Figure out order
        ordered_statements = eval_sort(statements_dependent, set(reserved_names | set(current_state["names"])))

        for k, v in current_state["grid"].items():
            if not Path(k).is_dir():
                logging.exception(f"Grid folder {k} does not exist")
                exceptions.append(FileNotFoundError(f"No such file or directory: {repr(k)}"))
                continue
            try:
                stack = v["stack"]
                values = eval_all(ordered_statements, v["stack"])
                stack.update({statement.name: v for statement, v in zip(ordered_statements, values)})
                write_grid(k, stack, files_static, files_grid, self.root)
            except Exception as e:
                exceptions.append(e)
        if len(exceptions) > 0:
            raise exceptions[-1]

    def run_exec(self):
        """
        Performs the run action.

        Executes a command in all grid folders.
        """
        logging.info(f"Executing {' '.join(self.extra)}")
        current_state = self.load_state()
        logging.info(f"   grid folders: {len(current_state['grid'])}")
        exceptions = []
        for cwd in current_state["grid"]:
            try:
                print(f'{cwd}: {" ".join(self.extra)}')
                print(subprocess.check_output(self.extra, cwd=cwd, stderr=subprocess.PIPE, text=True))

            except FileNotFoundError as e:
                print(f"{' '.join(self.extra)}: file not found (working directory {repr(cwd)})")
                logging.exception(f"{' '.join(self.extra)}: file not found (working directory {repr(cwd)})")
                exceptions.append(e)

            except subprocess.CalledProcessError as e:
                print(f"{' '.join(self.extra)}: process error (working directory {repr(cwd)})")
                print(e.stdout, end="")
                print(e.stderr, end="")
                logging.exception(f"{' '.join(self.extra)}: process error (working directory {repr(cwd)})")
                exceptions.append(e)
        if len(exceptions) > 0:
            raise exceptions[-1]

    def run_cleanup(self):
        """
        Performs the cleanup action.

        Removes all grid folders and grid state file.
        """
        logging.info("Cleaning up")
        current_state = self.load_state()
        logging.info("Removing grid folders")
        exceptions = []
        for f in current_state["grid"]:
            try:
                shutil.rmtree(f)
                logging.info(f"  {f}")
            except Exception as e:
                exceptions.append(e)
                logging.exception(f"Error while removing {f}")
        if len(exceptions):
            logging.error(f"{len(exceptions)} errors occurred while removing grid folders")
        logging.info("Removing the data file")
        Path(self.state_fn).unlink()
        if len(exceptions):
            raise exceptions[-1]

    def run(self):
        self.setup_logging()
        if self.action == "new":
            self.run_new()
        elif self.action == "distribute":
            self.run_distribute()
        elif self.action == "run":
            self.run_exec()
        elif self.action == "cleanup":
            self.run_cleanup()
        else:
            raise NotImplementedError(f"action '{self.action}' not implemented")


def grid_run(options=None):
    """Parses command line arguments and runs the desired grid action"""
    if options is None:
        options = arg_parser.parse_args()

    if options.action == "new":
        if len(options.extra) > 0:
            arg_parser.error("usage: grid new (with no extra arguments)")
    elif options.action == "run":
        if len(options.extra) == 0:
            arg_parser.error("usage: grid run COMMAND")
    elif options.action == "cleanup":
        if len(options.extra) > 0:
            arg_parser.error("usage: grid cleanup (no extra arguments)")
    elif options.action == "distribute":
        if len(options.extra) == 0:
            arg_parser.error("usage: grid distribute FILE [FILE ...]")

    return Engine.from_argparse(options).run()



def linspace(start, end, steps):
    """Like numpy.linspace"""
    assert steps > 2
    assert isinstance(steps, int)
    result = []
    for i in range(steps):
        w = 1.0 * i / (steps - 1)
        result.append(start * (1 - w) + end * w)
    return result


builtins = {
    "range": range,
    "linspace": linspace,
}
